#
# %W%
#
# code for tree structure in tcl
# data will be stored in array each of whose elements are lists
#  the list will contain the following data:
#  { nodetype parentid childid rtsiblingid {{attrname attrval} ...}
#    {textdata} }
# the array will be indexed starting at 1 so that in id==0 will
# represent the null ptr.
#
# variables:
# treedata      array containing all of the nodes for each of the trees listed by treeid
# rootid        array containing the root node for each of the trees indexed by treeid
# displaytreeid array containing the name of the displaytree for each treeid
#
# methods:
# int    getnewid{treeid}
# void   clean{}
# void   init{treeid}
# int    makenewroot{treeid nodetype attriblist textlist}
# int    makenewnode{treeid nodetype attriblist textlist}
# string getdisplaytreeid{treeid}
# void   setdisplaytreeid{treeid dtreeid}
# string getnodetype{id}
# void   setnodetype{id nodetype}
# int    getparentid{id}
# void   setparentid{id parentid}
# int    getchildid{id}
# void   setchildid{id childid}
# int    getsiblingid{id}
# void   setsiblingid{id siblingid}
# list   getattriblist{id}
# void   setattriblist{id attriblist}
# list   gettextdata{id}
# void   settextdata{id testlist}
# void   insert_rightmostchild{parentid newid}
# void   insert_beside{leftid newid}
# void   print_depth_tree{id}
# void   remove_node{id}
# void   prune_node{id}
# int    move_node{id newparentid}


namespace eval TreeData {
    namespace export init
    # contains the data for all of the trees it is an
    # array indexed by index.treeid where index is an integer
    #  and treeid is a unique string give to TreeData to 
    #  identifying a particular tree
    variable treedata

    # contains the rootnodes for each of the trees.  It is
    # indexed by treeid
    variable rootid

    # contains the name for the displaytree for a given treeid.
    variable displaytreeid

proc getnewid {treeid} {
    # finds the maximum id in use
    # returns the maximumid + 1
    variable treedata
    set maxindex 0

    set nodelist [array name treedata *.$treeid]
    if {$nodelist != ""} {
	foreach id $nodelist {
	    regexp (\[0-9\]+)\.$treeid $id tmp index
	    if {$index > $maxindex} {
		set maxindex $index
	    }
	}
    }
    incr maxindex
    return $maxindex.$treeid
}

proc init {treeid} {
    # initializes the data structure for a specific treeid
    set rootid($treeid) 0
    set displaytreeid($treeid) 0
    set treedata($treeid.0) 0
    return
}
    
    

proc clean {} {
    # cleans out the entire tree
    # should be only called once
    # returns nothing
    variable treedata
    variable rootid
    variable displaytreeid

    set nodelist [array names treedata]
    if {$nodelist != ""} {
	foreach i $nodelist {
	    unset treedata($i)
	}
    }
    set nodelist [array names rootid]
    if {$nodelist != ""} {
	foreach i $nodelist {
	    unset rootid($i)
	}
    }
    set nodelist [array names displaytreeid]
    if {$nodelist != ""} {
	foreach i $nodelist {
	    unset displaytreeid($i)
	}
    }
    array set treedata {}
    array set rootid {}
    array set displaytreeid {}
    return
}


proc makenewroot {treeid nodetype attriblist textlist} {
    # creates a new root using
    # the passed in data.
    variable treedata
    variable rootid
    set nodelist [array names treedata *.$treeid]
    if {$nodelist != ""} {
	foreach i $nodelist {
	    unset treedata($i)
	}
    }
    set rootid($treeid) [makenewnode $treeid $nodetype $attriblist $textlist]
    return $rootid($treeid)
}

proc makenewnode {treeid nodetype attriblist textlist} {
    # creates new node using the passed in data
    # return the id of the new node
    variable treedata
    set newid [getnewid $treeid]
    set treedata($newid) [list $nodetype 0 0 0 $attriblist $textlist]

    return $newid
}

proc getrootid {treeid} {
    variable rootid
    return $rootid($treeid)
}

proc getdisplaytreeid {treeid} {
    variable displaytreeid
    return $displaytreeid($treeid)
}

proc setdisplaytreeid {treeid dtreeid} {
    variable displaytreeid
    set displaytreeid($treeid) $dtreeid
    return
}

proc getnodetype {id} {
    # given an id returns the nodetype
    variable treedata
    return [lindex $treedata($id) 0]
}

proc setnodetype {id type} {
    # given a nodeid and a datavalue sets the nodetype
    variable treedata
    set parentid [lindex $treedata($id) 1]
    set childid  [lindex $treedata($id) 2]
    set siblingid [lindex $treedata($id) 3]
    set attriblist [lindex $treedata($id) 4]
    set textlist [lindex $treedata($id) 5]
    set treedata($id) [list $type $parentid $childid $siblingid $attriblist $textlist]
    return
}

proc getparentid {id} {
    # give the nodeid returns the parentid
    variable treedata
    return [lindex $treedata($id) 1]
}

proc setparentid {id p_id} {
    # given a nodeid and a datavalue sets the parentid
    variable treedata
    set nodetype [lindex $treedata($id) 0]
    set childid  [lindex $treedata($id) 2]
    set siblingid [lindex $treedata($id) 3]
    set attriblist [lindex $treedata($id) 4]
    set textlist [lindex $treedata($id) 5]
    set treedata($id) [list $nodetype $p_id $childid $siblingid $attriblist $textlist]
    return
}

proc getchildid {id} {
    # give the nodeid returns the childid
    variable treedata
    return [lindex $treedata($id) 2]
}

proc setchildid {id c_id} {
    # given a node_id and a datavalue sets the childid
    variable treedata
    set nodetype [lindex $treedata($id) 0]
    set parentid [lindex $treedata($id) 1]
    set siblingid [lindex $treedata($id) 3]
    set attriblist [lindex $treedata($id) 4]
    set textlist [lindex $treedata($id) 5]
    set treedata($id) [list $nodetype $parentid $c_id $siblingid $attriblist $textlist]
    return
}

proc getsiblingid {id} {
    # give the nodeid returns the parentid
    variable treedata
    return [lindex $treedata($id) 3]
}

proc setsiblingid {id s_id} {
    # given a node_id and a datavalue sets the rtsiblingid
    variable treedata
    set nodetype [lindex $treedata($id) 0]
    set parentid [lindex $treedata($id) 1]
    set childid  [lindex $treedata($id) 2]
    set attriblist [lindex $treedata($id) 4]
    set textlist [lindex $treedata($id) 5]
    set treedata($id) [list $nodetype $parentid $childid $s_id $attriblist $textlist]
    return
}

proc getattriblist {id} {
    # give the nodeid returns the parentid
    variable treedata
    return [lindex $treedata($id) 4]
}

proc setattriblist {id attributes} {
    # given a node_id and a datavalue sets the attriblist
    variable treedata
    set nodetype [lindex $treedata($id) 0]
    set parentid [lindex $treedata($id) 1]
    set childid  [lindex $treedata($id) 2]
    set siblingid [lindex $treedata($id) 3]
    set textlist [lindex $treedata($id) 5]
    set treedata($id) [list $nodetype $parentid $childid $siblingid $attributes $textlist]
    return
}

proc gettextdata {id} {
    # give the nodeid returns the textlist
    variable treedata
    return [lindex $treedata($id) 5]
}

proc settextdata {id textdata} {
    # given a node_id and a datavalue sets the textlist
    variable treedata
    set nodetype [lindex $treedata($id) 0]
    set parentid [lindex $treedata($id) 1]
    set childid  [lindex $treedata($id) 2]
    set siblingid [lindex $treedata($id) 3]
    set attriblist [lindex $treedata($id) 4]
    set treedata($id) [list $nodetype $parentid $childid $siblingid $attriblist $textdata]
    return
}

proc insert_rightmostchild {parentid newid} {
    # insert new node into the rightmost child slot beneath parent
    # check if parent has children
    variable treedata
    if {[getchildid $parentid] == 0} {
	# no children
	setchildid $parentid $newid
	setparentid $newid $parentid
    } else {
	# children
	set currentid [getchildid $parentid]
	while {[getsiblingid $currentid] != 0} {
	    set currentid [getsiblingid $currentid]
	}
	setparentid $newid $parentid
	setsiblingid $newid 0
	setsiblingid $currentid $newid
    }
    return
}

proc insert_besidechild {left new} {
    # inserts new node to the right of the left node
    # as long as the id's aren't the same.
    if {$left != $new && [getsiblingid $left] != $new} {
	set leftsibling [getsiblingid $left]
	setsiblingid $new $leftsibling
	setsiblingid $left $new
	setparentid $new [getparentid $left]
    }
    return
}

proc insert_firstchild {parent newchild} {
    # inserts a node into the first child position
    set originalchild [getchildid $parent]
    if {$newchild != $originalchild} {
	setsiblingid $newchild $originalchild
	setparentid $newchild $parent
	setchildid $parent $newchild
    }
}

proc print_depth_tree {id} {
    # does a depth first printing of the tree rooted at id
    puts "[getnodetype $id] -- [gettextdata $id]"
    if {[getchildid $id] != 0} {
	# children
	set current [getchildid $id]
	while {$current != 0} {
	    print_depth_tree $current
	    set current [getsiblingid $current]
	}
    }
    return
}

proc remove_node {id} {
    # removes the node and the subtree beneath it
    
    variable treedata
    # Am I the direct child of parent?
    if {[getchildid [getparentid $id]] == $id} {
	# if so make my sibling the child of parent
	setchildid [getparentid $id] [getsiblingid $id]
    } else {
	# if not, then I am in the middle
	# so, find my left sibling by looping
	set leftsibling [getchildid [getparentid $id]]
	while {[getsiblingid $leftsibling] != $id} {
	    set leftsibling [getsiblingid $leftsibling]
	}
	# now fix pointers
	setsiblingid $leftsibling [getsiblingid $id]
    }
    # I am now longer a child of my parent
    # now destroy my children if I have any.
    # I think I have to get rid of my children from
    # the right working leftwards or else I am left
    # with dangling data.
    # Get ids of children in reverse order
    set child [getchildid $id]
    if {$child != 0} {
	set children $child
	while {[getsiblingid $child] != 0} {
	    set child [getsiblingid $child]
	    set children [linsert $children 0 $child]
	}
	#puts "Removing $children"
	foreach i $children {
	    remove_node $i
	}
    }
    # now that my children are gone delete myself
    unset treedata($id)
    return
}

proc prune_node {id} {
    # removes the children from id but keeps node id.
    
    variable treedata

    # find my children and remove them from right to left
    set child [getchildid $id]
    if {$child != 0} {
	set children $child
	while {[getsiblingid $child] != 0} {
	    set child [getsiblingid $child]
	    set children [linsert $children 0 $child]
	}
	foreach i $children {
	    remove_node $i
	}
    }
    return
}

proc move_node {id newparentid {leftid 0}} {
    # moves the subtree rooted at id to the node, newparentid,
    # and with its left sibling being leftid.  If leftid is
    # not specified it will default to 0 meaning we should do
    # a firstchild insertion.
    # must check to see that the current node does not lie
    # along the upward inheritance chain of the newparentid
    # if it does disallow move.
    # We also need to check that leftid is not id
    # Note: this does not check that the newparentid and
    #       leftid have a parent/child relationship!

    if {$id == 0} {
	#puts "id is 0"
	return
    }
    if {$newparentid == 0} {
	#puts "newparentid is 0"
	return
    }
    if {$id == $leftid} {
	#puts "leftid is the same as id"
	return
    }
    if {$id == $newparentid} {
	#puts "Cannot be own parent"
	return
    }

    set current $newparentid
    while {[getparentid $current] != 0} {
	if {[getparentid $current] == $id} {
	    # error moving node will have itself as an ancestor!
	    return 1
	}
	set current [getparentid $current]
    }
    # new parent does not have the moving node as an ancestor
    # we continue by performing the same steps as in remove node
    # but we do not delete the node.

    # is this the node the first child of its parent
    if {[getchildid [getparentid $id]] == $id} {
	# if so make the sibling the first child of parent
	setchildid [getparentid $id] [getsiblingid $id]
    } else {
	# if not, then I am in the middle
	# so, find my left sibling by looping
	set leftsibling [getchildid [getparentid $id]]
	while {[getsiblingid $leftsibling] != $id} {
	    set leftsibling [getsiblingid $leftsibling]
	}
	# now fix pointers
	setsiblingid $leftsibling [getsiblingid $id]
    }

    # clear pointers but not the child pointer
    #setparentid $id 0
    #setsiblingid $id 0

    # move the node and its children
    if {$leftid == 0} {
	# do a firstchild insertion of node
	insert_firstchild $newparentid $id
    } else {
	# do a beside insertion using leftid
	insert_besidechild $leftid $id
    }
}


}
