# @(#)moving_node.tcl	1.6
# Contains the code for moving a tree's node from one place to another.


proc setup_move_bindings {canvas} {
    # bindings associated with the selected node that allows it to be moved
    $canvas bind selected <Button-3> "create_rect $canvas %x %y"
    $canvas bind selected <Button3-Motion> "move_rect $canvas %x %y"
    $canvas bind selected <ButtonRelease-3> "drop_rect $canvas %x %y"
}

# globals needed to preserved the current x and y coords for proper motion of the rectangle
set prior_x 0
set prior_y 0


proc create_rect {canvas x y} {
    # create a rectangle the same size and location as the selected node's 
    # bbox
    global prior_x
    global prior_y

    set prior_x $x
    set prior_y $y

    set nodeid [lindex [$canvas gettags selected] 1]
    scan [$canvas bbox $nodeid:box] "%f %f %f %f" x1 y1 x2 y2
    set id [$canvas create rectangle $x1 $y1 $x2 $y2 -tags temp_rect]
}

proc move_rect {canvas x y} {
    global prior_x
    global prior_y

    set dx [expr $x - $prior_x]
    set dy [expr $y - $prior_y]

    $canvas move temp_rect $dx $dy
    set prior_x $x
    set prior_y $y
}


proc drop_rect {canvas x y} {
    # 1. if this is the first node then don't check for neighbors link the node to root
    # 2. if there already are nodes then do proximity checks.
    # find where the node is dropped
    # check for surrounding nodes and then link
    # to the closest one.

    # To determine the closest node we shall go through a multistep process
    # Step 1 - Find new parent
    #      1a - if overlapping find nearest above us that will be new parent
    #      1b - if no overlapping find nearest above us that will be new parent
    #      1c - if none above us link to DN2 node
    # Step 2 - Find our place in list of children
    # Step 3 - Move children from end of list until our new place to a holding location
    # Step 4 - Move our node to new parent
    # Step 5 - Move moved children back to parent

    # Errors to check for: 1. can't move top dn2 node
    #                      2. respond gracefully to $tree move errors
    #   Consider checking for allowable_subnodes

    global allowable_subnode
    set tree [TreeData::getdisplaytreeid $canvas]
    set found_closest 0
    set is_vertical_tree [expr [string compare [$tree cget -layout] "vertical"] == 0]
    scan [$canvas bbox temp_rect] "%f %f %f %f" bbx1 bby1 bbx2 bby2
    set centerx [expr ($bbx1 + $bbx2) / 2.0]
    set centery [expr ($bby1 + $bby2) / 2.0]
    
    set overlapping_objects [$canvas find overlapping $bbx1 $bby1 $bbx2 $bby2]
    if {[llength $overlapping_objects] > 1} {
	# the temp_rect does overlap something
	# remember to check for non-node objects
	# cull from the list anything that is not a node or is below us or is us or is to our right
	set overlapping_nodes {}
	foreach i $overlapping_objects {
	    scan [$canvas bbox $i] "%f %f %f %f" tmpx1 tmpy1 tmpx2 tmpy2
	    set tags [$canvas gettags $i]
	    if {$is_vertical_tree} {
		# vertical tree
		# condition looks weird but y coord starts in upper left corner
		# so for below y coord must be greater than another
		# tests that center of temp_rect is below overlapping object and the overlapping
		# object is a node and that the overlapping is not the node that is moving
		if { ($centery > [expr ($tmpy1 + $tmpy2) / 2.0]) && [lsearch -regexp $tags node] != -1 &&
		[string compare [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $i] 1]] } {
		    # overlapping rect is above temp rect
		    lappend overlapping_nodes $i
		}
	    } else {
		# horizontal tree
		if { ($centerx < [expr ($tmpx1 + $tmpx2) / 2.0]) && [lsearch -regexp $tags node] != -1 &&
		[string compare [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $i] 1]] } {
		    # overlapping rect is right of temp rect
		    lappend overlapping_nodes $i
		}
	    }
	}
	# now find the closest node
	if {[llength $overlapping_nodes] > 0} {
	    # we have overlapping nodes above us so find closest
	    find_closest $canvas centerx centery overlapping_nodes closest_id found_closest
	}
	if {$found_closest} {
	    set nodetype_to_move [lindex [$canvas gettags selected] 0]
	    set potential_parenttype [lindex [$canvas gettags $closest_id] 0]
	    if {[lsearch $allowable_subnode([string toupper $potential_parenttype]) [string toupper $nodetype_to_move]] == -1} {
		# closest overlapping node is not an appropriate parent
		# so reset found_closest flag and try closest node above or leftward of dropped rectangle
		set found_closest 0
	    }
	}
    }
    if {!$found_closest} {
	# since nothing overlapped us then search for node above us or for horizontal tree to left
	scan [$canvas cget -scrollregion] "%d %d %d %d" origin_x origin_y max_x max_y
	if {$is_vertical_tree} {
	    set intersecting_objects [$canvas find overlapping $origin_x $origin_y $max_x $bby1]
	} else {
	    set intersecting_objects [$canvas find overlapping $origin_x $origin_y $bbx1 $max_y]
	}
	if {[llength $intersecting_objects] > 1} {
	    # there exists objects above us
	    # clean out any non-node objects
	    set above_nodes {}
	    foreach i $intersecting_objects {
		set tags [$canvas gettags $i]
		# tests that overlapping object is a node and not the same object as being moved
		if {[lsearch -regexp $tags node] != -1 && [string compare [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $i] 1]] } {
		    lappend above_nodes $i
		}
	    }
	    if {[llength $above_nodes] > 0} {
		find_closest $canvas centerx centery above_nodes closest_id found_closest
	    }
	}
    }
    if {$found_closest} {
	#we have found the closest node now do some checking to see if attachment is legal
	#puts "Found closest!!!:   id=$closest_id tags=[$canvas gettags $closest_id]"
	set nodetype_to_move [lindex [$canvas gettags selected] 0]
	set potential_parenttype [lindex [$canvas gettags $closest_id] 0]
	if {[lsearch $allowable_subnode([string toupper $potential_parenttype]) [string toupper $nodetype_to_move]] != -1} {
	    # nodetype of node to move can be attached to potential_parent

	    # - simple algorithm; find first child at which their vertical centerline is to the right of me
	    #   this is the child before which I should be entered.

	    # adding brief test for current parent == potential parent if this is true
	    # move the node_to_move to the root first then get the children of the parent.
	    # This allows us to reorganize the order of the children of a parent
	    #if {![string compare [lindex [$canvas gettags $closest_id] 1] [$tree parent [lindex [$canvas gettags selected] 1]]]} {
		# keeping same parent
		#$tree move [lindex [$canvas gettags selected] 1] {}
	    #}
	    array set children {}
	    get_children_nodes $canvas [lindex [$canvas gettags $closest_id] 1] children
	    if {[array size children] == 0} {
		# no children
		TreeData::move_node [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $closest_id] 1]
	    } else {
		# there are children so figure out where to insert the moving node
		set done 0
		set i 0
	    
		if {$is_vertical_tree} {
		    # tree is vertical
		    # find last child with its vertical centerline to the left of my centerline
		    while {!$done && $i < [array size children]} {
			set box [lindex $children($i) 1]
			set testcenterx [expr ([lindex $box 0] + [lindex $box 2])/2.0]
			if {$testcenterx > $centerx} {
			    set done 1
			} else {
			    incr i
			}
		    }
		    set i [expr $i - 1]
		} else {
		    # tree is horizontal
		    # find last child with its horizontal centerline above my centerline
		    while {!$done && $i < [array size children]} {
			set box [lindex $children($i) 1]
			set testcentery [expr ([lindex $box 1] + [lindex $box 3])/2.0]
			if {$testcentery > $centery} {
			    set done 1
			} else {
			    incr i
			}
		    }
		    set i [expr $i - 1]
		}
		#puts "Moving selected node"
		if {$i == -1} {
		    # the moving node will become the first child 
		    #puts "Moving to first: [lindex [$canvas gettags selected] 1]"
		    #puts "\tparent: [lindex [$canvas gettags $closest_id] 1]"
		    TreeData::move_node [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $closest_id] 1]
		} else {
		    #puts "Moving: [lindex [$canvas gettags selected] 1]"
		    #puts "\tparent: [lindex [$canvas gettags $closest_id] 1]"
		    #puts "\tleft: [lindex $children($i) 0]"
		    TreeData::move_node [lindex [$canvas gettags selected] 1] [lindex [$canvas gettags $closest_id] 1] [lindex $children($i) 0]
		}
	    }
	} else {
	    tk_messageBox -title "Cannot Move Node" -message \
	     "Node $nodetype_to_move cannot be attached to $potential_parenttype"
	}
    } else {
	puts "Haven't found closest.  I think we should quit!"
    }
    $canvas delete temp_rect
    MakeViewableTree $canvas
}

proc find_closest {canvas x y nodes id found_it} {
    upvar 1 $x centerx
    upvar 1 $y centery
    upvar 1 $nodes list_of_nodes
    upvar 1 $id closest_id
    upvar 1 $found_it found_closest

    set found_closest 0
    set olddist 10000000
    foreach i $list_of_nodes {
	scan [$canvas bbox $i] "%f %f %f %f" tmpx1 tmpy1 tmpx2 tmpy2
	set tmpcenterx [expr ($tmpx1 + $tmpx2) / 2.0]
	set tmpcentery [expr ($tmpy1 + $tmpy2) / 2.0]
	set newdist [expr pow(($centerx - $tmpcenterx),2) + pow(($centery - $tmpcentery),2)]
	if {$newdist < $olddist} {
	    set olddist $newdist
	    set closest_id $i
	    set found_closest 1
	}
    }
}

proc get_children_nodes {canvas node centers} {
    # receives a node id and returns an array containing the child node tags
    # and their bboxes
    upvar 1 $centers carray

    # empty the centers array
    foreach i [array names carray] {
	unset carray($i)
    }

    set cnt 0
    set current [TreeData::getchildid $node]
    if {$current != 0} {
	set carray($cnt) [list $current [$canvas bbox $current]]
	incr cnt
	while {[TreeData::getsiblingid $current] != 0} {
	    set current [TreeData::getsiblingid $current]
	    set carray($cnt) [list $current [$canvas bbox $current]]
	    incr cnt
	}
    }

    #foreach i [$tree subnodes $node] {
    #	set carray($cnt) [list [$canvas gettags $i] [$canvas bbox $i]]
    #	incr cnt
    #}
}


